from __future__ import annotations

from importlib.metadata import version

from dahlia.lib import Dahlia, Depth

TEST_STRING = "&R".join(f"&{c * 2}" for c in "0123456789abcdefhijklmno")
VERSION = version("dahlia")

if __name__ == "__main__":
    print(f"Dahlia v{VERSION}")
    if (max_depth := Dahlia().depth) is None:
        print("Disabled colors")
    else:
        print(f"Max depth: {max_depth.name} ({max_depth.value}-bit)")
        for depth in Depth.__members__.values():
            Dahlia(depth=depth).print(TEST_STRING)
            if depth is max_depth:
                break



import pytest

from dahlia.__main__ import TEST_STRING
from dahlia.utils import clean, clean_ansi, escape


@pytest.mark.parametrize(
    ("content", "marker", "expected"),
    [
        ("&e&nunderlined&rn yellow", "&", "underlined yellow"),
        ("&e&nunderlined&rn yellow", "!", "&e&nunderlined&rn yellow"),
        ("!e!nunderlined!rn yellow", "!", "underlined yellow"),
        ("§_4 gives §4red", "§", "§4 gives red"),
    ],
)
def test_clean(content: str, marker: str, expected: str) -> None:
    assert clean(content, marker) == expected


@pytest.mark.parametrize(
    ("content", "expected"),
    [
        ("\x1b[93m\x1b[4munderlined\x1b[0m yellow", "underlined yellow"),
        ("\x1b[38;2;255;255;85m\x1b[4munderlined\x1b[0m yellow", "underlined yellow"),
        ("\x1bxxx", "\x1bxxx"),
        ("\x1b[xm", "\x1b[xm"),
    ],
)
def test_clean_ansi(content: str, expected: str) -> None:
    assert clean_ansi(content) == expected


def test_cli_test_string() -> None:
    assert clean(TEST_STRING) == "0123456789abcdefhijklmno"


@pytest.mark.parametrize(
    ("content", "marker", "expected"),
    [
        ("&e&nunderlined&rn yellow", "&", "&_e&_nunderlined&_rn yellow"),
        ("&ame & &dher", "&", "&_ame &_ &_dher"),
        ("&e&nunderlined&rn yellow", "!", "&e&nunderlined&rn yellow"),
        ("!e!nunderlined!rn yellow", "!", "!_e!_nunderlined!_rn yellow"),
        ("§_4 gives §4red", "§", "§__4 gives §_4red"),
    ],
)
def test_escape(content: str, marker: str, expected: str) -> None:
    assert escape(content, marker) == expected



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from dahlia.constants import (
    ANSI_REGEX,
    CODE_REGEXES,
    NO_GROUP_CODES,
    REGEX_BREAKING_MARKERS,
)

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator


def clean(string: str, marker: str = "&") -> str:
    """Removes all Dahlia codes from a string."""
    for code, *_ in _find_codes(string, _with_marker(marker)):
        string = string.replace(code, marker if code[1] == "_" else "")
    return string


def clean_ansi(string: str) -> str:
    """Removes all ANSI codes from a string."""
    for ansi_code in _find_ansi_codes(string):
        string = string.replace(ansi_code, "")
    return string


def escape(string: str, marker: str = "&") -> str:
    """Escapes all instances of the marker in a string."""
    return string.replace(marker, marker + "_")


def _find_codes(
    string: str, patterns: Iterable[re.Pattern[str]]
) -> Iterator[tuple[str, bool | None, str]]:
    return reversed(
        dict.fromkeys(
            (match[0], None, match[1])  # full-reset, escape
            if (full_snd_char := match[0][1]) in NO_GROUP_CODES
            else (
                (match[0], None, match[1])  # reset
                if full_snd_char == "r"
                else (match[0], match[1] == "~", match[2])  # regular
            )
            for pattern in patterns
            for match in pattern.finditer(string)
        )
    )


def _find_ansi_codes(string: str) -> set[str]:
    return {match[0] for match in ANSI_REGEX.finditer(string)}


def _with_marker(marker: str) -> list[re.Pattern[str]]:
    if not isinstance(marker, str):
        msg = "The marker has to be a string"
        raise TypeError(msg)
    if len(marker) != 1:
        msg = "The marker has to be a single character"
        raise ValueError(msg)
    if marker in REGEX_BREAKING_MARKERS:
        marker = "\\" + marker
    return [re.compile(marker + i) for i in CODE_REGEXES]



from __future__ import annotations

from enum import Enum
from os import getenv
from typing import Any, Literal, cast

from dahlia.constants import (
    BG_FORMAT_TEMPLATES,
    COLOR_SETS,
    FORMAT_TEMPLATES,
    FORMATTERS,
    RESET,
)
from dahlia.utils import _find_codes, _with_marker, clean

DepthInt = Literal[3, 4, 8, 24]


class Depth(Enum):
    """Specifies usable color depth levels."""

    TTY = 3
    """3-bit color (tty)"""
    LOW = 4
    """4-bit color"""
    MEDIUM = 8
    """8-bit color"""
    HIGH = 24
    """24-bit color (true color)"""


class Dahlia:
    """The main Dahlia class handling string transformations."""

    __slots__ = (
        "_auto_reset",
        "_depth",
        "_hash_fields",
        "_marker",
        "_no_color",
        "_patterns",
        "_reset",
    )

    _depth: DepthInt | None

    def __init__(
        self,
        *,
        depth: Depth | DepthInt | str | None = None,
        marker: str = "&",
        auto_reset: bool = True,
    ) -> None:
        self._no_color = bool(getenv("NO_COLOR")) or getenv("TERM") == "dumb"
        if depth is None:
            self._depth = None if self._no_color else _resolve_depth().value
        elif isinstance(depth, int):
            self._depth = Depth(depth).value
        elif isinstance(depth, str):
            self._depth = Depth[depth.upper()].value
        else:
            self._depth = depth.value
        self._marker = marker
        self._auto_reset = auto_reset
        self._patterns = _with_marker(marker)
        self._reset = marker + "R"
        self._hash_fields = (self._depth, self._auto_reset, self._marker)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Dahlia):
            return NotImplemented
        return self._hash_fields == other._hash_fields

    def __hash__(self) -> int:
        return hash(self._hash_fields)

    def __repr__(self) -> str:
        return (
            f"Dahlia(depth={self._depth}, "
            f"auto_reset={self._auto_reset}, marker={self._marker!r})"
        )

    @property
    def depth(self) -> Depth | None:
        """Specifies what ANSI color set to use (in bits)."""
        return self._depth and Depth(self._depth)

    @property
    def marker(self) -> str:
        """Specifies the prefix used by format codes ("&" by default)."""
        return self._marker

    @property
    def auto_reset(self) -> bool:
        """When True, appends a full reset code to the transformed string."""
        return self._auto_reset

    def convert(self, string: str) -> str:
        """Transforms a Dahlia string to an ANSI string."""
        if self._no_color:
            return clean(string)
        if self.auto_reset and not string.endswith(self._reset):
            string += self._reset
        for code, bg, color in _find_codes(string, self._patterns):
            string = string.replace(code, self._get_ansi(color, bg=bg))
        return string.replace(self._marker + "_", self._marker)

    def input(self, prompt: str) -> str:
        """Wraps the built-in `input` by transforming the prompt."""
        return input(self.convert(prompt))

    def print(self, *args: object, **kwargs: Any) -> None:
        """Wraps the built-in `print` by transforming all of its args."""
        print(*map(self.convert, map(str, args)), **kwargs)

    def _get_ansi(self, code: str, *, bg: bool | None) -> str:
        if code == "R":
            return "\033[0m"
        if code == "_":
            return self._marker + code
        if code[0] == "r":
            return f"\033[{RESET[code[1]]}m"
        formats = BG_FORMAT_TEMPLATES if bg else FORMAT_TEMPLATES
        code_size, rem3 = divmod(len(code), 3)
        if not rem3:
            return formats[24].format(
                ";".join(
                    (str(int(code[i : i + 2], 16)) for i in (0, 2, 4))
                    if code_size == 2
                    else (str(int(i, 16) * 0x11) for i in code)
                )
            )
        if code in FORMATTERS:
            return formats[3].format(FORMATTERS[code])

        depth = cast(int, self._depth)
        value = COLOR_SETS[depth][code]
        if bg and depth <= 4:
            return formats[depth].format(int(value) + 10)
        return formats[depth].format(value)


def _resolve_depth() -> Depth:
    if getenv("COLORTERM") in {"truecolor", "24bit"}:
        return Depth.HIGH
    if (
        (term := getenv("TERM", "")) in {"terminator", "mosh"}
        or "24bit" in term
        or "24-bit" in term
    ):
        return Depth.HIGH
    if "256" in term:
        return Depth.MEDIUM
    return Depth.LOW



from unittest.mock import patch

import pytest

from dahlia.constants import REGEX_BREAKING_MARKERS
from dahlia.lib import Dahlia, Depth


@pytest.mark.parametrize(
    ("depth", "string", "expected"),
    [
        (Depth.HIGH, "foo", "foo"),
        (
            Depth.LOW,
            "&e&nunderlined&rn yellow",
            "\x1b[93m\x1b[4munderlined\x1b[24m yellow",
        ),
        (Depth.MEDIUM, "&7ball", "\x1b[38;5;248mball"),
        (Depth.HIGH, "&7ball", "\x1b[38;2;170;170;170mball"),
        (Depth.TTY, "&kcarcinogen", "\x1b[5mcarcinogen"),
        (Depth.LOW, "&lhyperextension", "\x1b[1mhyperextension"),
        (Depth.MEDIUM, "&hgammopathy", "\x1b[8mgammopathy"),
        (Depth.LOW, "&asupremacism", "\x1b[92msupremacism"),
        (Depth.HIGH, "&mlymphocyte", "\x1b[9mlymphocyte"),
        (Depth.TTY, "&alingualumina", "\x1b[32mlingualumina"),
        (Depth.TTY, "&3filariidae", "\x1b[36mfilariidae"),
        (Depth.LOW, "&4aura", "\x1b[31maura"),
        (Depth.LOW, "&~4aura", "\x1b[41maura"),
        (Depth.TTY, "&r&R", "&r\x1b[0m"),
        (Depth.HIGH, "&9aspergillaceae", "\x1b[38;2;85;85;255maspergillaceae"),
        (Depth.LOW, "&~edemigod", "\x1b[103mdemigod"),
        (Depth.TTY, "&9miller", "\x1b[34mmiller"),
        (Depth.HIGH, "&#ffaff3;gleaming", "\x1b[38;2;255;175;243mgleaming"),
        (Depth.HIGH, "&#fbf;gweaming", "\x1b[38;2;255;187;255mgweaming"),
    ],
)
def test_conversion(depth: Depth, string: str, expected: str) -> None:
    assert Dahlia(depth=depth, auto_reset=False).convert(string) == expected


@pytest.mark.parametrize(
    ("marker", "expected"),
    [
        ("&", "\x1b[93me§ee§§_4x"),
        ("e", "&\x1b[93m§\x1b[93m§§_4x"),
        ("§", "&ee\x1b[93me§§4x"),
        ("_", "&ee§ee§§\x1b[31mx"),
        ("4", "&ee§ee§§_4x"),
        ("x", "&ee§ee§§_4x"),
    ],
)
def test_marker_clashing(marker: str, expected: str) -> None:
    assert (
        Dahlia(marker=marker, auto_reset=False, depth=Depth.LOW).convert("&ee§ee§§_4x")
        == expected
    )


@pytest.mark.parametrize("marker", REGEX_BREAKING_MARKERS)
def test_regex_markers(marker: str) -> None:
    assert (
        Dahlia(depth=Depth.LOW, marker=marker).convert(f"{marker}4xe5")
        == "\x1b[31mxe5\x1b[0m"
    )


@pytest.mark.parametrize("marker", ["", "&&"])
def test_invalid_marker(marker: str) -> None:
    with pytest.raises(ValueError, match="The marker has to be a single character"):
        Dahlia(marker=marker)


@pytest.mark.parametrize("marker", [123, [1], {"key": "value"}])
def test_nonstring_marker(marker: str) -> None:
    with pytest.raises(TypeError, match="The marker has to be a string"):
        Dahlia(marker=marker)


@pytest.mark.parametrize(("auto_reset", "expected"), [(True, "\x1b[0m"), (False, "")])
def test_auto_reset(auto_reset: bool, expected: str) -> None:
    assert Dahlia(auto_reset=auto_reset).convert("") == expected


def test_print(capsys: pytest.CaptureFixture[str]) -> None:
    d = Dahlia()
    content = "&e&nunderlined&rn yellow"
    d.print(content)

    assert capsys.readouterr().out == d.convert(content) + "\n"


def test_print_custom_sep_end(capsys: pytest.CaptureFixture[str]) -> None:
    content = ("&#ffaff3;gleaming", "&bdiamond", "&lcool")
    d = Dahlia()
    d.print(*content, sep=":::", end="!!!")

    assert capsys.readouterr().out == ":::".join(d.convert(c) for c in content) + "!!!"


def test_input(capsys: pytest.CaptureFixture[str]) -> None:
    d = Dahlia(depth=Depth.LOW)
    prompt = "&a&lprompt: "
    with patch(
        "builtins.input", side_effect=lambda _: (print(d.convert(prompt)), "ok")[1]
    ):
        ans = d.input(prompt)

    assert capsys.readouterr().out == d.convert(prompt) + "\n"
    assert ans == "ok"


def test_depth_accepts_case_insensitive_str() -> None:
    assert Dahlia(depth="lOw").depth is Depth.LOW


def test_properties() -> None:
    d = Dahlia(depth=Depth.TTY)
    assert d.marker == "&"
    assert hash(d) == hash(Dahlia(depth=Depth.TTY))
    assert eval(repr(d)) == d  # noqa: S307
    assert d != 1



from __future__ import annotations

from unittest.mock import patch

import pytest

from dahlia.lib import Dahlia, Depth


@pytest.mark.parametrize(
    ("term", "colorterm", "expected"),
    [
        ("terminator", "foobar", Depth.HIGH),
        ("mosh", "foobar", Depth.HIGH),
        ("xterm-24bit", None, Depth.HIGH),
        (None, None, Depth.LOW),
        ("xterm-24-bit", None, Depth.HIGH),
        ("mosh", "truecolor", Depth.HIGH),
        ("vt", "foobar", Depth.LOW),
        (None, "truecolor", Depth.HIGH),
        ("xterm-ghostty", "24bit", Depth.HIGH),
        ("dumb", "foobar", None),
        ("vt", None, Depth.LOW),
        ("xterm-256color", None, Depth.MEDIUM),
        ("xterm-256color", "truecolor", Depth.HIGH),
    ],
)
def test_depth_resolution(
    term: str | None, colorterm: str | None, expected: Depth
) -> None:
    env = {k: v for k, v in (("TERM", term), ("COLORTERM", colorterm)) if v is not None}
    with patch("dahlia.lib.getenv", env.get):
        assert Dahlia().depth is expected


@pytest.mark.parametrize(
    "env",
    [
        {"TERM": "dumb"},
        {"NO_COLOR": "1"},
        {"NO_COLOR": "true"},
        {"NO_COLOR": "0"},
    ],
)
def test_dumb_term_is_no_color(env: dict[str, str]) -> None:
    with patch("os.environ", env):
        assert Dahlia().convert("&3x") == "x"


@pytest.mark.parametrize(
    "env",
    [
        {"NO_COLOR": ""},
        {"TERM": "some-term"},
        {},
    ],
)
def test_no_color_not_handled(env: dict[str, str]) -> None:
    with patch("os.environ", env):
        assert Dahlia(depth=3).convert("&3x") == "\x1b[36mx\x1b[0m"
